import jd2xx.JD2XX;
import jd2xx.JD2XXInputStream;
import jd2xx.JD2XXOutputStream;
import java.io.IOException;
import java.lang.Thread;

public class ConnectionCheck {
	
	public static JD2XX ftdi;
	
	/**
	 * Lists all connected USB FT2232 devices; exits the program if none are found.
	 * Connects to the first available device and sets it up to use the MPSSE.
	 */
	public static void listDevicesAndConnect() throws IOException {
		Object[] devs = null;
		devs = ftdi.listDevicesByDescription();
		
		if(devs.length < 1) {
			System.err.println("No DLP2232M device found!");
			System.exit(1);
		}
		
		System.out.println("Found DLP2232M devices:");
		for (int i=0; i<devs.length; ++i)
			System.out.println("\t(" + i + ") " + devs[i]);
		
		ftdi.open(0);
		System.out.println("Connected to device 0.");
		
		ftdi.resetDevice();
		System.out.println("Device reset.");
		
		ftdi.setLatencyTimer(5);
		ftdi.setBitMode(0x00, 0x00);
		ftdi.setBitMode(0x00, 0x02);
		setClockSpeed100k();
		syncMPSSE();
	}
	
	/**
	 * Clears the transmit and receive buffers, and sends an invalid command to make sure
	 * that we get an error like we expect.  If something is wrong, the program exits after
	 * displaying an error message.
	 */
	public static void syncMPSSE() throws IOException {
		byte syncResponse[] = new byte[2];
		
		ftdi.purge(JD2XX.PURGE_RX | JD2XX.PURGE_TX);
		
		// Send a bad command to ensure that MPSSE is synchronized
		ftdi.write(0xAA);
		ftdi.read(syncResponse);
		if( (0xFF & (int)syncResponse[0]) != 0xFA || (0xFF & (int)syncResponse[1]) != 0xAA ) {
			System.out.println("MPSSE is not synchronized");
			System.exit(1);
		} else {
			System.out.println("MPSSE is synchronized");
		}
		
	}
	
	/**
	 * Sets the speed of the SPI clock to 100 KHz. (See page 19 of FTDI Application Note 108)
	 */
	public static void setClockSpeed100k() throws IOException {
		byte[] dataOut = new byte[3];
		dataOut[0] = (byte)0x86;  // Set clock divisor
		dataOut[1] = (byte)0x78;  // (Clock divisor-1) low byte
		dataOut[2] = (byte)0x00;  // (Clock divisor-1) high byte
		ftdi.write(dataOut);
	}
	
	/**
	 * This does some simple communication with the microcontroller, one byte at a time.
	 * You'll want to replace this function with your own that implements your
	 * specific communications protocol.
	 */
	public static int communicateNoProtocol(int message) throws IOException {
		byte[] dataOut = new byte[10];
		byte[] dataIn = new byte[1];
		
		// The FTDI chip's MPSSE works by sending it a string of bytes containing commands
		// (and their arguments.)  See the FTDI Application Note 108 which describes the meaning
		// of all of these commands and their arguments.
	
		// Toggle chip select low
		dataOut[0] = (byte)0x80; // setup PORT
		dataOut[1] = (byte)0x00; // Set CS low
		dataOut[2] = (byte)0x0B; // SK, DO, CS are outputs, DI, GPIOL1-L4 inputs
		
		// Send and receive the actual SPI data to and from the microcontroller
		dataOut[3] = (byte)0x35; // clock data out on negative edge and in on negative edge, MSB first
		dataOut[4] = (byte)0x00; // low byte of (length-1)
		dataOut[5] = (byte)0x00; // high byte of (length-1)
		dataOut[6] = (byte)message; // payload
		
		// Toggle chip select high
		dataOut[7] = (byte)0x80; // setup PORT
		dataOut[8] = (byte)0x08; // Set CS high
		dataOut[9] = (byte)0x0B; // SK, DO, CS are outputs, DI, GPIOL1-L4 inputs
		ftdi.write(dataOut);
		ftdi.read(dataIn);
		return 0xFF & (int)dataIn[0];
	}
	
	public static void main(String[] args) throws Exception {
		ftdi = new JD2XX();	
		listDevicesAndConnect();
		
		// If we make it to this point, we're connected, synchronized, and ready to start
		// communicating over SPI.
		System.out.println("Device is ready!");
		
		// Send 10 bytes to the microcontroller and print out the response
		for(int i=0; i<10; i++) {
			int response = communicateNoProtocol(i);
			System.out.println("Sent " + i + " and received " + response);
			Thread.sleep(1000);
		}
		
	}
	
}